package ch.epfl.collections;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;

/**
 * Generic and iterable implementation of the Set interface. The underlying
 * implementation is using an unbalanced binary search tree.
 * 
 * For the tree to work with class E, E has to implement Comparable interface,
 * which provides us with .compareTo() method. ( see Node.add() ). This allows
 * us to manage any class E as we did plain integers in the first part of the
 * exercise (i.e. the items can be compared between themselves; they have
 * total ordering)
 * 
 * For explanation about the meaning of public methods of TreeSet, please
 * see the Set interface in the same package.
 * 
 * @see Set
 *
 */
public class TreeSet<E extends Comparable<E>> implements Set<E> {
	/** initially the root is a Leaf element, no sub-elements */
    private Tree<E> root = new Leaf<E>();

    @Override
    public boolean isEmpty() {
        return root.isEmpty();
    }

    @Override
    public int size() {
        return root.size();
    }

    @Override
    public void add(E newElem) {
        root = root.add(newElem);
    }

    @Override
    public void remove(E elem) {
        root = root.remove(elem);
    }

    @Override
    public boolean contains(E elem) {
        return root.contains(elem);
    }

    /**
     * Returns the anonymous class - iterator of items in the tree.
     * The iterator is based on a stack memory structure - last item
     * to be added to the stack will be the first one to be removed from it
     * (also known as LIFO structure). In our case, it means that the last
     * element added to the stack will be the first element the iterator will
     * return.
     * 
     * Not all nodes of the tree are pushed onto the stack at once. The iterator
     * relies on Node method .stackSmallerNodes() which traverse the tree until it
     * finds the smallest element and adds all elements on the way to the smallest.
     * 
     * Say this is our tree with 7 nodes (Leaf elements are NOT shown).
     * 
     *    <- smaller  elements   greater ->
     *                   A
     *                  /  \
     *                B     C
     *               /\    / \
     *              D  E  F   G
     *              
     * Initial call to the iterator() will push nodes [ A, B, D ] to the stack, D being
     * at the end of the stack and first one to be returned by the iterator.
     * Nodes [ E, C, F, G ] are not on the stack.
     * 
     * When calling the next() method of the iterator it will:
     * 	1) return node D, remove it from the stack
     *  2) try to push the greater child of D, which won't do much - D has no node children
     *     and Leaf nodes are not iterated over
     *     
     * Stack now contains [ A, B ].
     * 
     * The second call to next() will:
     *  1) return node B, remove it from the stack
     *  2) push greater child of B, which is in this case E
     *  
     * Stack now contains [ A, E ].
     * 
     * ... and so on....
     * 
     * So, what happens when we call next() and we have only [ A ] on stack?
     *  1) return node A, remove it from the stack
     *  2) put greater child of A, which is in this case C, but also all smaller children
     *     which is in this case only F
     * 
     * Stack now contains [ C, F ].
     * 
     * The advantages of this approach are:
     *  1) the items are returned in the correct order - from the smallest to the biggest.
     *     For this example, the nodes are returned by the iterator in the order
     *     D, B, E, A, F, C, G
     *     which corresponds to the general order
     *     D < B < E < A < F < C < G
     *  2) stack has a nicer memory consumption
     * 
     * @see Node#stackSmallerNodes(LinkedList)
     */
    @Override
    public Iterator<E> iterator() {
        return new Iterator<E>() {
            LinkedList<Node<E>> nodeStack = new LinkedList<Node<E>>();
            { root.stackSmallerNodes(nodeStack); } /** this code is put in the anonymous class' constructor */

            public boolean hasNext() {
                return !nodeStack.isEmpty();
            }

            public E next() {
                if (nodeStack.isEmpty())
                    throw new NoSuchElementException();
                Node<E> nextNode = nodeStack.pop(); /** get current node, and remove it from stack */
                nextNode.g.stackSmallerNodes(nodeStack); /** pay attention to the g !!!! */
                return nextNode.v;
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

    @Override
    public String toString() {
        StringBuilder b = new StringBuilder("[");
        Iterator<E> it = iterator();
        while (it.hasNext()) {
            b.append(it.next());
            if (it.hasNext())
                b.append(",");
        }
        return b.append("]").toString();
    }
    
    /**
     * Tree interface for internal usage of TreeSet.
     * 
     * Interface that is implemented by both Node and Leaf elements of the tree.
     *
     */
    private static interface Tree<E extends Comparable<E>> {
    	/**
    	 * Is the current node empty. It basically allows us to tell
    	 * the difference between a Leaf and a Node (Nodes are always
    	 * full).
    	 * 
    	 * @return <code>true</code> if tree element has a value.
    	 */
        public boolean isEmpty();
        
        /**
         * Returns a number of nodes under the current tree-element (including that
         * element). Leafs count as zero (i.e. are not taken into account, since they
         * don't contain any real value).
         * 
         * @return int size of the (sub)tree.
         * 
         * @see Node#size()
         */
        public int size();
        
        /**
         * Adds an element under the current node.
         * @param newElem New element to be added.
         * @return The new Node element containing the item.
         * 
         * @see Leaf#add(Comparable)
         */
        public Tree<E> add(E newElem);
        
        /**
         * Removes the element from the current node.
         * 
         * @param elem Element value to be removed
         * @return New root of the (sub)tree.
         * 
         * @see Node#remove(Comparable)
         */
        public Tree<E> remove(E elem);
        
        /**
         * Checks if the element containing value elem is contained as a subnode of
         * the current tree.
         * 
         * @param elem
         * @return <code>true</code> if there exists a node containing value elem.
         */
        public boolean contains(E elem);
        
        /**
         * Returns the smallest element of the (sub)tree.
         * 
         * @see Node#getMin()
         * @return Value of the smallest node E.
         */
        public E getMin();
        
        /**
         * Pushes all the smaller sub-nodes which have value
         * (i.e. no Leaf elements are pushed) to the nodeStack.
         * Used for the iterator() method.
         * 
         * @param nodeStack Stack into which we're pushing the nodes.
         * 
         * @see TreeSet#iterator()
         */
        public void stackSmallerNodes(LinkedList<Node<E>> nodeStack);
    }

    /**
     * Node is the tree-node which contains an value (of type E).
     *
     * @param <E>
     */
    private static class Node<E extends Comparable<E>> implements Tree<E> {
        private final E v; /** value of the current node */
        private Tree<E> s, g; /** s == smaller; g == greater */

        /**
         * The constructor is typically called from the {@link Leaf#add(Comparable)}.
         * The reason for this is that a new node will be always added at a position
         * previously occupied by the Leaf, replacing that Leaf.
         * 
         * @param s Smaller sub-tree.
         * @param e This node's value.
         * @param g Greater sub-tree.
         */
        public Node(Tree<E> s, E e, Tree<E> g) {
            this.s = s;
            this.v = e;
            this.g = g;
        }

        @Override
        public boolean isEmpty() {
            return false;
        }

        /**
         * The size of the tree is calculated recursively calling the size
         * of each sub-tree (greater sub-tree and smaller sub-tree) and adding
         * 1 for the current node. Leaf are not contributing to the size of the
         * tree, since they don't contain an actual value, and are a helper
         * structure.
         * 
         * @see Tree#size()
         * @see Leaf#size()
         */
        @Override
        public int size() {
            return 1 + s.size() + g.size();
        }

        /**
         * Adds the element in the sub-tree. We are handling 3 possibilities:
         *  - compared to the current element, the value is greater - put it in
         *    the g sub-tree
         *  - compared to the current element, the value is smaller - put it in
         *    the s sub-tree
         *  - the value is equal to the current node - we can ignore the new
         *    element, since Set needs to store a value only once (no duplicate
         *    elements)
         */
        @Override
        public Tree<E> add(E newElem) {
            int c = newElem.compareTo(v);
            if (c < 0)
                s = s.add(newElem);
            else if (c > 0)
                g = g.add(newElem);
            else
                ; // element already in set: do nothing
            return this;
        }

        /**
         * Removes the element from the tree, and reorganizes the remaining
         * items into a updated tree structure.
         * 
         * While we haven't found the value elem to be removed, we traverse the tree
         * in the greater or smaller direction.
         * 
         * The most complicated situation occurs when we reach the node that we want
         * to remove, but it has both "smaller" and "greater" sub-tree
         * (i.e. both s and g are Nodes, not Leafs). In order for the tree to stay valid,
         * we need to replace the current node (the one being removed) with the smallest
         * node of the g sub-tree (we call it "newE"), which we can retrieve with g.getMin().
         * After we replace the node being removed with the value "newE",
         * we also have to remove it from the g sub-tree, so we wouldn't have duplicated
         * nodes in the tree. For this reason we call g.remove(newE).
         */
        @Override
        public Tree<E> remove(E elem) {
            int c = elem.compareTo(v);
            /** Traverse the tree until the element is found. */
            if (c < 0) {
                s = s.remove(elem);
                return this;
            } else if (c > 0) {
                g = g.remove(elem);
                return this;
            } else {
            	/** We found the element to be removed. */
                if (s.isEmpty()) return g; /** if we don't have smaller element, just use g subtree node to replace current node being removed */
                else if (g.isEmpty()) return s; /** likewise if we don't have greater element, just use s subtree node to replace current node being removed */
                else { /** complicated case - the node is replaced by the smallest node of the g sub-tree */
                    E newE = g.getMin();
                    return new Node<E>(s, newE, g.remove(newE)); /** notice the g.remove(newE) call which becomes the new g sub-tree */
                }
            }
        }

        @Override
        public boolean contains(E elem) {
            int c = elem.compareTo(v);
            if (c < 0)
                return s.contains(elem);
            else if (c > 0)
                return g.contains(elem);
            else
                return true;
        }

        @Override
        public E getMin() {
            return s.isEmpty() ? v : s.getMin();
        }

        @Override
        public void stackSmallerNodes(LinkedList<Node<E>> nodeStack) {
            nodeStack.push(this);
            s.stackSmallerNodes(nodeStack);
        }
    }

    /**
     * Leaf is a placeholder for the new elements in the tree.
     * 
     * Most operations return dummy responses and contain logic which keeps
     * the algorithm in check. The most significant operation is the 
     * add() method which is responsible for creating new Nodes when the
     * end of the tree is reached.
     *
     * @param <E>
     */
    private static class Leaf<E extends Comparable<E>> implements Tree<E> {
        @Override
        public boolean isEmpty() {
            return true;
        }

        @Override
        public int size() {
            return 0;
        }

        @Override
        public Tree<E> add(E newElem) {
            return new Node<E>(new Leaf<E>(), newElem, new Leaf<E>());
        }

        @Override
        public Tree<E> remove(E elem) {
            return this;
        }

        @Override
        public boolean contains(E elem) {
            return false;
        }

        @Override
        public E getMin() {
            throw new UnsupportedOperationException();
        }

        @Override
        public void stackSmallerNodes(LinkedList<Node<E>> nodeStack) {
            // nothing to do
        }
    }
}
